import json, time
from time import sleep
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import *
from threading import Thread

#constant variables.
URL = "https://www.trading212.com/trading-instruments/cfd"#
COOKIES = '//*[@id="__next"]/main/div/div/div[2]/div[2]/div[2]/div[1]'
LIST_FILED = '//*[@id="__next"]/main/section[2]/div/div/div[2]/div/div/div[1]/div[1]/div'
LIST_TITLES = '//*[@id="__next"]/main/section[2]/div/div/div[1]/div[2]'
SCROLL_AREA = '//*[@id="__next"]/main/section[2]/div/div/div[2]/div/div/div[2]/div'
EXPAND_TRIGGER = '//*[@id="__next"]/main/section[2]/div/div/div[2]/div/div/div[1]/div[2]/div[1]/div'
SEARCH_FIELD = '//*[@id="__next"]/main/section[2]/div/div/div[1]/div[1]/div[2]/div/div[1]/input'
SCROLL_AREA = '//*[@id="__next"]/main/section[2]/div/div/div[2]/div/div/div[1]/div[1]/div/div' #STYLE = height: 29520px; width: 100%;
SCROLL_TOP = '//*[@id="__next"]/main/section[2]/div/div/div[2]/div/div/div[2]/div' #STYLE height: 53.7805px; top: 0px; get top value

def setup(search_value):
    options = webdriver.ChromeOptions()
    options.headless=True
    driver = webdriver.Chrome(options=options)
    driver.get(URL)
    driver.fullscreen_window()
    driver.implicitly_wait(5)
    try: driver.find_element(By.XPATH,f"{COOKIES}").click()
    except Exception as em: return f"exception caught: {type(em): {em}}"
    try: driver.find_element(By.XPATH,f"{SEARCH_FIELD}").send_keys(f"{search_value}")
    except Exception as em: print(f"exception caught: {type(em): {em}}")
    sleep(2)
    return driver
   
def get_data(search_value, sleep_time: float = 0.01):
    scrollinto = "arguments[0].scrollIntoView();"
    set_keys = {"UK":{}, "INDEX":{}, "NASDAQ":{},
                "NYSE":{}, "COMMODITIES":{}, "SWITZERLAND":{},
                "HONG KONG":{}, "OTC":{}, "FINLAND":{},
                "DENMARK":{}, "IRELAND":{}, "AUSTRIA":{},
                "PORTUGAL":{}, "RUSSIAN":{}, "SWEDEN":{},
                "POLAND":{}, "BELGIUM":{}, "NORWAY":{},
                "NETHERLANDS":{}, "SPAIN":{}, "AUSTRALIA":{},
                "FRENCE":{}, "ITALY":{}, "CANADA":{},
                "DEUTSCHE CFD":{}}
    
    if search_value not in set_keys.keys(): return f"{search_value} is not a valued arg"
    
    driver = setup(search_value)
    #Creating json file
    try:
        with open('CFD_list.json','r+',encoding=('utf-8')) as f:
            file_data = json.load(f)
            for v in set_keys:
                if v not in file_data:
                    file_data.update({f"{v}":{}})
            f.seek(0)
            json.dump(file_data,f,indent=4)
    except FileNotFoundError:
        with open('CFD_list.json','w',encoding=('utf-8')) as f:
            json.dump(set_keys,f,indent=4)

    #Save scrapped data to json file.
    
    #Get data area size
    scroll_area = driver.find_element(By.XPATH,f"{SCROLL_AREA}").get_attribute("Style")
    scroll_value = scroll_area.split(';')
    num = ''.join(c for c in scroll_value[0] if c not in 'height:px ')

    while True:
        t1 = time.perf_counter()
        try: scrapped = driver.find_elements(By.XPATH,
            f'{LIST_FILED}/descendant::div[@class="item-wrapper"]')
        except NoSuchElementException as em: break
        #Check scrapped data and end when no more new data is avalible
        try:
            if last_set == scrapped: break
        except UnboundLocalError as em:
            pass
        #Update document
        for v in scrapped:
            try: items = v.text.split('\n')
            except Exception as em: pass
            avg_spread = ''.join(c for c in items[-1] if c not in '-,£')
            if avg_spread: avg_spread = float(avg_spread)
            if not avg_spread: avg_spread = "N/A"
            try: data = {f"{items[-5]}":
                         {"name": f"{items[-6]}",
                          "isin": f"{items[-4]}",
                          "avg spread": avg_spread}}
            except IndexError as em: pass
            #Write to new file 
            with open('CFD_list.json','r+',encoding=('utf-8')) as f:
                file_data = json.load(f)
                main_key = file_data[f"{search_value}"]
                if items[-5] not in main_key:
                    file_data[f"{search_value}"].update(data)
                    f.seek(0)
                    json.dump(file_data, f, indent=4)

            last_set = scrapped
            try:
                driver.implicitly_wait(5)
                driver.execute_script(scrollinto, v)
            except StaleElementReferenceException as em:
                pass

        t2 = time.perf_counter()
    driver.close()
    return "Scrap completed"



# this unofficial API was created by Flock92 originally made to automate my CFD trading on the trading212 platform

import requests
from selenium import webdriver
from selenium.webdriver.common.by import By
from getpass import getpass
from .apitconstant import *
from time import sleep
import logging
import json


class Apit212:
    headers = {}

    def __init__(
            self,
            username: str,
            password: str,
            timeout: int = 2,
            interval: float = 5.0,
            mode: str = 'demo',
            headers: dict = None):
        """login to t212 account to get credentials to run API

        :param username: trading212 account username
        :type username: str
        :param password: trading212 account password
        :type password: str
        :param timeout: login attempt limits
        :type timeout: int
        :param interval: adjust for slow internet connections
        :type interval: float
        :param mode: set to 'demo' by default
        :type interval: str
        """

        # START LOGGING
        logging.basicConfig(filename="apit212.log",
                            format='%(asctime)s :: %(levelname)s :: %(message)s',
                            filemode='w')

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        self.logger.info(f'timeout: {timeout}, interval: {interval}, mode: {mode}')

        # CHECK PARAM'S
        if isinstance(username, str):
            pass
        else:
            username = None

        if isinstance(password, str):
            pass
        else:
            password = None

        if isinstance(timeout, int):
            pass
        else:
            raise TypeError(':param timeout: invalid input')

        if isinstance(interval, (int, float)):
            pass
        else:
            raise TypeError(':param interval: invalid input')

        if isinstance(mode, str) and mode in ['demo', 'live']:
            pass
        else:
            raise TypeError(':param mode: invalid input :: expected "demo" or "live" ')

        # set variables for functions
        self.url = f'https://{mode}.trading212.com/'
        options = webdriver.FirefoxOptions()
        options.add_argument('--log-level=3')
        options.headless = True
        my_cookie = None

        # Get login details if no information was passed.
        if username is None:
            username = input('username: ')
        elif password is None:
            password = getpass('password: ')
        else:
            pass

        # STARTUP WEBDRIVER
        d = webdriver.Firefox(options=options)

        try:
            d.get(url=URL)
            self.logger.info(msg=f'successfully loaded {URL}')
        except Exception as em:
            self.logger.error(em)

        # CHECK URL
        if d.current_url != URL:
            self.logger.error("can not verify URL"), quit(d.close())
        else:
            pass

        d.implicitly_wait(30)

        # START LOGIN PROCESS
        while True:
            if isinstance(headers, dict) or timeout == 0:
                break
            else:
                pass
            timeout -= 1
            try:
                d.find_element(By.XPATH, f"{COOKIE_POPUP}").click()
                self.logger.info('cookie popup closed')
            except Exception as em:
                self.logger.error(f'failed to close cookies popup: {em}')
                continue
            sleep(interval)
            try:
                d.find_element(By.XPATH, f"{LOGIN_BUTTON}").click()
                self.logger.info('login form opened')
            except Exception as em:
                self.logger.error(f'failed to open login form: {em}')
                continue

            # INPUT DETAILS
            try:
                d.find_element(By.NAME, "email").send_keys(username)
                self.logger.info('input username')
            except Exception as em:
                self.logger.error(f'failed to input username: {em}')
                continue
            try:
                d.find_element(By.NAME, "password").send_keys(password)
                self.logger.info('input password')
            except Exception as em:
                self.logger.error(f'failed to input password: {em}')
                continue
            try:
                d.find_element(By.XPATH, f"{SUBMIT_BUTTON}").click()
                self.logger.info('form submitted')
            except Exception as em:
                self.logger.error(f'failed submit form: {em}')
                continue
            sleep(interval)
            try:
                d.execute_script("arguments[0].click();", d.find_element(By.XPATH, f"{POPUP_CLOSE}"))
                self.logger.info('popup closed')
            except Exception as em:
                self.logger.warning(f'failed to close popup: {em}')
                continue
            try:
                user_name = d.find_element(By.XPATH, f"{USER_NAME}").text
                self.logger.info('username verified')
            except Exception as em:
                self.logger.error(f'failed to verify username: {em}')
                continue

            # VERIFY LOGIN
            if username in user_name:
                self.logger.info('successfully logged into account.')
                break
            else:
                self.logger.warning(f'failed logging attempt :: {timeout}')
                pass

        # SWITCH ACCOUNT
        try:
            d.get(f"https://{mode}.trading212.com")
            self.logger.info(f'switched to "{mode}"')
        except Exception as em:
            self.logger.error(f'failed to switch to {mode}: {em}')

        sleep(interval)

        # GET COOKIES
        user_agent = d.execute_script("return navigator.userAgent;")

        if isinstance(headers, dict):
            self.headers = headers
        else:
            cookies = d.get_cookies()
            for cookie in cookies:
                if cookie['name'] == f'TRADING212_SESSION_{mode.upper()}':
                    my_cookie = f"TRADING212_SESSION_{mode.upper()}={cookie['value']};"
                else:
                    pass

            self.headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": user_agent,
                "Cookie": f'{my_cookie}',
            }

        self.logger.info(f'finished setup...{username}')

    # RETURNS HEADERS AND URL
    def __getitem__(self, key):
        return self.headers

    # GET AUTH VALIDATE
    def auth_validate(self) -> dict:
        """

        :return: {'id': '********-****-****-****-************', 'accountId': ********, 'customerId': ********,
        'tradingType': 'CFD', 'customerUuid': '********-****-****-****-************', 'frontend': 'WC4',
        'readyToTrade': True, 'deviceUuid': ''}
        """
        r = requests.get(f'{self.url}/auth/validate', headers=self.headers)
        return r.json()

    # GET ACCOUNT DETAILS
    def get_account(self) -> dict:
        """Get account info

        :return: {'demoAccounts': [{'id': ********, 'type': 'DEMO', 'tradingType': 'CFD',
        'customerId': ********, 'createdDate': '2023-01-17T03:20:48.000+00:00',
        'status': 'ACTIVE', 'registerSource': 'WC4', 'currencyCode': 'GBP', 'readyToTrade': True}],
        'liveAccounts': [{'id': ********, 'type': 'LIVE', 'tradingType': 'CFD', 'customerId': ********,
        'createdDate': '2023-01-17T03:20:32.000+00:00', 'status': 'PENDING_ACTIVATION', 'registerSource': 'WC4',
        'currencyCode': 'GBP', 'readyToTrade': False}]}
        """
        r = requests.get(f'{self.url}/rest/v1/accounts', headers=self.headers)
        return r.json()

    # GET ACCOUNT FUNDS
    def get_funds(self) -> dict:
        """Get account funds

        :return: dict={'20434246': {'accountId': ********,
        'tradingType': 'CFD', 'currency': 'GBP', 'freeForWithdraw': 310.5,
        'freeForCfdTransfer': 0, 'total': 4954.12, 'lockedCash': {'totalLockedCash': 0, 'lockedCash': []}}}
        """
        r = requests.get(f"{self.url}/rest/v2/customer/accounts/funds", headers=self.headers)
        return r.json()

    # GET ORDER SIZE
    def get_max_min(self, instrument) -> dict:
        """Get the min and max 'BUY' & 'SELL' for an instrument passed to this function.
        :param instrument:
        :return: {'minBuy': 1.0, 'maxBuy': 4593.17,
        'minSell': 1.0, 'maxSell': 0.0, 'sellThreshold': 0.0, 'maxSellQuantity': 0}
        """
        params = {'instrumentCode': {instrument}}
        r = requests.get(f"{self.url}/v1/equity/value-order/min-max",
                         headers=self.headers,
                         params=params)
        return r.json()

    # CANCEL ORDER
    def cancel_order(self, order_id) -> dict:
        """

        :param order_id:
        :return: {'account': {'dealer': 'AVUSUK', 'positions': [{'positionId': '********-****-****-****-************',
        'humanId': '**********', 'created': '2023-06-05T21:32:45.000+03:00', 'modified': None, 'averagePrice': 181.09,
        'averagePriceConverted': 144.96174516, 'currentPrice': 186.45, 'limitPrice': None, 'stopPrice': None,
        'value': 16047.62, 'investment': 15945.79, 'limitStopNotify': None, 'trailingStop': None,
        'trailingStopPrice': None, 'trailingStopNotify': None, 'code': 'AAPL', 'margin': 3225.65, 'ppl': 461.33,
        'quantity': 110.0, 'maxBuy': None, 'maxSell': None, 'maxOpenBuy': None, 'maxOpenSell': None, 'swap': -16.11,
        'frontend': 'WC4', 'pplAdjustment': None, 'autoInvestQuantity': None, 'fxPpl': None}
        """
        payload = {'positionId': f'{order_id}'}
        r = requests.delete(url=f"{self.url}/rest/v2/pending-orders/entry/{order_id}",
                            headers=self.headers, data=json.dumps(payload))
        return r.json()
    
    # CANCEL ALL PENDING ORDERS
    def cancel_all_orders(self) -> dict:
        """"""
        payload = []
        data = requests.post(url=f"{self.url}/rest/trading/v1/accounts/summary",
                             headers=self.headers, data=json.dumps(payload))

        r = requests.delete(url=f"{self.url}/rest/v2/pending-orders/cancel",
                            headers=self.headers, data=data)

        return r.json()

    # CANCEL ORDER
    def close_position(self, position_id, quantity, current_price) -> dict:
        """
        close open positions
        :param position_id:
        :param quantity:
        :param current_price:
        :return:
        """

        payload = {'coeff': {
            'positionId': f'{position_id}',
            'quantity': quantity,
            'targetPrice': current_price}}

        r = requests.delete(url=f"{self.url}/rest/v2/trading/open-positions/close/{position_id}",
                            headers=self.headers, data=json.dumps(payload))
        return r.json()

    # GET ACCOUNT SUMMARY (GET POSITION ID)
    def get_summary(self) -> dict:
        """This function can be used along with a bot to get positionId's to manually close positions and get data on
        your accounts.

        :return: {'cash': {'free': 219.73, 'total': 4851.04, 'interest': -379.02, 'indicator': 51, 'commission': 0.0,
        'cash': 5000.0, 'ppl': 327.56, 'result': -97.5, 'margin': 4631.31, 'spreadBack': 0.0, 'nonRefundable': 0.0,
        'dividend': 0.0, 'totalCashForWithdraw': 219.73}, 'open': {'unfilteredCount': 3,
        'items': [{'positionId': '********-****-****-****-************', 'humanId': '**********',
        'created': '2023-06-05T21:32:45.000+03:00', 'averagePrice': 181.09,
        'averagePriceConverted': 144.96174516, 'currentPrice': 186.31, 'value': 16013.4,
        'investment': 15945.79, 'code': 'AAPL', 'margin': 3218.77, 'ppl': 448.66,
        'quantity': 110.0, 'swap': -16.08, 'frontend': 'WC4'}
        """
        payload = []

        r = requests.post(url=f"{self.url}/rest/trading/v1/accounts/summary", headers=self.headers,
                          data=json.dumps(payload))

        return r.json()

    # GET LIST OF COMPANIES
    def get_companies(self) -> list:
        """
        This function will return a list of tickers and their corresponding isin code.
        :return: [{'ticker': 'TICK', 'isin': '*************'}]
        """

        r = requests.get(f"{self.url}/rest/companies", headers=self.headers)

        return r.json()

    # GET TICKER INFORMATION
    def get_instruments_info(self, instrument: str) -> dict:
        """
        This function returns information about the ticker.
        :param instrument:
        :return: {'code': 'TSLA', 'type': 'STOCK', 'margin': 0.2,
        'shortPositionSwap': -0.06510720836211, 'longPositionSwap': -0.25863029163789,
        'tsSwapCharges': '1970-01-01T23:00:00.000+02:00', 'marginPercent': '20', 'leverage': '1:5'}
        """
        r = requests.get(f'{self.url}/v2/instruments/additional-info/{instrument}',
                         headers=self.headers)

        return r.json()

    # GET INSTRUMENT INFORMATION
    def get_order_info(self, instrument: str, quantity):
        """

        :param instrument:
        :param quantity:
        :return:
        """
        params = {'instrumentCode': f"{instrument}",
                  'quantity': quantity,
                  'positionId': 'null'}

        r = requests.get(f"{self.url}/rest/v1/tradingAdditionalInfo", headers=self.headers, params=params)

        return r.json()

    # GET ASK PRICE FOR INSTRUMENT
    def get_ask(self, instrument: str, _useaskprice: str = "false") -> list:
        """

        :param instrument:
        :param _useaskprice:
        :return: [{'request': {'ticker': '****', 'useAskPrice': False}, 'response':
        {'timestamp': 1687852810000, 'price': 250.37, 'period': 'd1'}}]
        """

        payload = [{"ticker": f"{instrument}", "useAskPrice": f"{_useaskprice}"}]

        r = requests.put(f'{self.url}charting/v1/watchlist/batch/deviations',
                         headers=self.headers, data=json.dumps(payload))

        return r.json()

    # PLACE LIMIT ORDER
    def limit_order(self,
                    instrument: str,
                    quantity: float,
                    target_price: float,
                    take_profit: float,
                    stop_loss: float,
                    notify: str = "NONE"):
        """the minimum requirement to submit a limit order is the instrument, quantity and target price.
        You can also set a take_profit and stop_loss limits. I would like to point out that this
        function will only submit your order and will make no attempts to change any of the parameters.

        :param instrument:
        :param quantity:
        :param target_price:
        :param take_profit:
        :param stop_loss:
        :param notify:
        :return:
        """
        payload = {'quantity': round(quantity, 1),
                   'targetPrice': round(target_price, 2),
                   'takeProfit': round(take_profit, 2),
                   'stopLoss': round(stop_loss, 2),
                   'notify': notify}

        r = requests.post(f'{self.url}/rest/v2/pending-orders/entry-dep-limit-stop/{instrument}',
                          headers=self.headers,
                          data=json.dumps(payload))

        return r.json()

    # MARKET ORDER
    def market_order(self,
                     instrument: str,
                     target_price: float,
                     quantity: int,
                     limit_distance: float,
                     stop_distance: int,
                     notify: str = "NONE"):
        """the minimum requirement to submit a market order is the instrument, quantity and target price.
        You can also set a take_profit and stop_loss limits. I would like to point out that this
        function will only submit your order and will make no attempts to change any of the parameters.

        if an incorrect order is submitted you will get a code message returned.
        ie: {'code': 'BusinessException', 'context': {'max': 684, 'type': 'InsufficientFundsMaxBuy'},
        'message': 'InsufficientResourcesException'}

        :param instrument:
        :param target_price:
        :param limit_distance:
        :param notify:
        :param quantity:
        :param stop_distance:
        :return:
        """

        payload = {'instrumentCode': f"{instrument}",
                   'limitDistance': limit_distance,
                   'notify': f"{notify}",
                   'quantity': quantity,
                   'stopDistance': stop_distance,
                   'targetPrice': target_price}

        r = requests.post(url=f"{self.url}/rest/v2/trading/open-positions",
                          headers=self.headers, data=json.dumps(payload))

        return r.json()

    # ADD TRAILING STOP LOSS
    def trailing_stop(self, position_id: str, distance: float, notify: str = "NONE"):
        """
        Add a trailing stop to you live positions.
        :param position_id:
        :param distance:
        :param notify:
        :return:
        """

        payload = {"ts": {"distance": distance}, "notify": f"{notify}"}

        r = requests.put(f"{self.url}/rest/v2/pending-orders/associated/{position_id}",
                         headers=self.headers, data=json.dumps(payload))

        return r.json()

    # RESET DEMO ACCOUNT
    def _reset(self, account_id: int, amount: int, currency_code: str):
        """"""
        payload = {"accountId": account_id, "amount": amount, "currencyCode": f"{currency_code}", "reason": "settings"}

        r = requests.post(f"{self.url}/rest/v1/account/reset-with-sum", headers=self.headers, data=json.dumps(payload))

        return r



#CONSTANT ELEMENTS FROM apit212
URL = 'https://www.trading212.com/'
LOGIN_BUTTON = '//*[@id="__next"]/header/div/div/div[2]/div[1]/div[2]/p'
COOKIE_POPUP = '//*[@id="__next"]/main/div[2]/div/div[2]/div[2]/div[2]/div[1]'
SUBMIT_BUTTON = '//*[@id="__next"]/main/div[2]/div/div[2]/div/div[2]/div/form/div[5]/input'
POPUP_CLOSE = '//*[@id="app"]/div[6]/div/div[2]/div/div/div/div[2]/div/div[1]/div[3]/div/span/div'
REMEBER_ME_TOGGLE = '//*[@id="__next"]/main/div[2]/div/div[2]/div/div[2]/div/form/div[4]/div[2]/input'
USER_NAME = '//*[@id="app"]/div[5]/div[1]/div/div[2]/div[3]/div/div[1]/div[1]'

#GET COOKIES
COOKIE_KEYS = ('TRADING212_SESSION_LIVE', 'TRADING212_SESSION_DEMO','LOGIN_TOKEN')

#ACCOUNT STATUS
ACCOUNT_VALUE = '//*[@id="app"]/div[3]/div[1]/div/div[2]/div[1]/div/span[2]'
FREE_FUNDS = '//*[@id="app"]/div[3]/div[2]/div[2]/div/div[2]/div[2]/div[4]/div[2]/div[2]'
LIVE_RESULTS = '//*[@id="app"]/div[3]/div[2]/div[2]/div/div[2]/div[2]/div[4]/div[1]/div[2]'
BLOCKED_FUNDS = '//*[@id="app"]/div[3]/div[2]/div[2]/div/div[2]/div[2]/div[4]/div[3]/div[2]'
MARGIN_INDICATORS = '//*[@id="app"]/div[5]/div[2]/div[2]/div/div[2]/div[2]/div[4]/div[4]/div/div/div/div'

APP = '//*[@id="app"]'

SETUP_ERR = 'FAILED TO SETUP PROFILE'
